// -*- C++ -*-
#ifndef GENERATOR_HH
#define GENERATOR_HH 

#include "AGILe/AGILe.hh"
#include "AGILe/Utils.hh"
#include "AGILe/Status.hh"
#include "AGILe/Particle.hh"
#include "AGILe/GeneratorState.hh"
#include "HepMC/GenEvent.h"

// Temporary forward declaration until this exists...
//class GeneratorState;

namespace AGILe {

  /// @author Andy Buckley
  class Generator {
  public:

    /// Standard constructor.
    Generator() :
    _initialstateset(false), _initialized(false),_versionSet(false) { }
    
    /// Destructor.
    virtual ~Generator() { }

    /// Define initial state.
    Status setInitialState(int p1 = PROTON, double e1 = 7000.0, 
                           int p2 = PROTON, double e2 = 7000.0){
      Status status = setGenSpecificInitialState(p1, e1, p2, e2);
      if(status==SUCCESS) _initialstateset = true;
      Beam beam1, beam2;
      beam1.name = p1;
      beam2.name = p2;
      beam1.energy = e1;
      beam2.energy = e2;
      beam1.position = 1;
      beam2.position = 2;
      _myState.setBeams(beam1, beam2);
      return status;
    };
    
    virtual const GeneratorState& getState(){
      string myName = getName();
      _myState.setName(myName);
      _myState.setVersion(getVersion());
      _myState.setCrossSection(getCrossSection());
      PDF pdf;
      for(vector<Beam>::const_iterator beamIt=_myState.getBeams().begin();
          beamIt != _myState.getBeams().end(); ++beamIt){
        pdf.particleId = beamIt->name;
        pdf.set = getPDFSet(beamIt->name);
        pdf.member = getPDFMember(beamIt->name);
        pdf.scheme = getPDFScheme(beamIt->name);
        _myState.addPDF(pdf);
      }
      return _myState;
      //return SUCCESS;
    }

    virtual Status setState(const GeneratorState& state) {
      throw runtime_error("This ain't ready yet, it's just here as an interface placeholder. Sorry.");
      return SUCCESS;
    }

    virtual const string getName() const {
      throw runtime_error("This shouldn't happen - Generator is a pure virtual base class which can't be instantiated.");
    }

    virtual Status setVersion(const string &version){
      _myVersion = version;
      _versionSet = true;
      return SUCCESS;
    }
    
    /// used as a last resort to determine the generator version.
    /// Ideally a specific generator will know its own version and overwrite 
    /// this.  Some generators (e.g. FHerwig) don't know this.
    virtual const string getVersion(){
      if(!_versionSet){
        cerr<<"You asked the generator "<<getName()
        <<" what version it is, but it does not know - "
        <<"you should set it with a call to setVersion(string)"<<endl;
        throw runtime_error("Version unknown for generator");
      }
      return _myVersion;
    }
    
    /// Get the PDF set used by the generator, e.g. HWLHAPDF (for herwig)
    /// or one of the default sets.
    virtual string getPDFSet(PdgCode pid){
      throw runtime_error("Have not implemented getPDFSet in " + getName());
    };
    /// Get the PDF id number for a given particle
    virtual int getPDFMember(PdgCode pid){
      throw runtime_error("Have not implemented getPDFMember in "+getName());
    };
    /// Get the PDF scheme used by the generator (LHAPDF or other)
    virtual string getPDFScheme(PdgCode pid)const{
      throw runtime_error("Have not implemented getPDFScheme in "+getName());
    };
    
    /// Set the random number generator seed.
    virtual Status setSeed(const int value) {
      if (_initialized) {
        cerr << "Can't set seed: generator has already been initialized." << endl;
        throw (runtime_error("Tried to set RNG seed after generator initialization."));
      }
      _myState.addSeed(value);

      return SUCCESS;
    }

    // Removing generic setParam since it doesn't play well with inheritance (it can't be virtual)
//     /// Pass a generic-typed parameter to the generator.
//     template <typename T>
//     Status setParam(const string& name, const T& value) {
//       if (_initialized) {
//         cerr << "Can't set parameter " << name << ": generator has already been initialized." << endl;
//         return FAILURE;
//       }
//       stringstream ss;
//       ss << value;
//       return setParam(name, ss.str());
//     }

    /// Pass an int-valued parameter to the generator.
    virtual Status setParam(const string& name, const int& value) {
      if (_initialized) {
        cerr << "Can't set parameter " << name << ": generator has already been initialized." << endl;
        throw (runtime_error("Tried to set param after generator initialization."));
      }
      return setParam(name, toString(value));
    }

    /// Pass an int-valued parameter to the generator.
    virtual Status setParam(const string& name, const unsigned int& value) {
      if (_initialized) {
        cerr << "Can't set parameter " << name << ": generator has already been initialized." << endl;
        throw (runtime_error("Tried to set param after generator initialization."));
      }
      return setParam(name, toString(value));
    }
    
    /// Pass an int-valued parameter to the generator.
    virtual Status setParam(const string& name, const long& value) {
      if (_initialized) {
        cerr << "Can't set parameter " << name << ": generator has already been initialized." << endl;
        throw (runtime_error("Tried to set param after generator initialization."));
      }
      return setParam(name, toString(value));
    }

    /// Pass an int-valued parameter to the generator.
    virtual Status setParam(const string& name, const unsigned long& value) {
      if (_initialized) {
        cerr << "Can't set parameter " << name << ": generator has already been initialized." << endl;
        throw (runtime_error("Tried to set param after generator initialization."));
      }
      return setParam(name, toString(value));
    }

    /// Pass a double/float-valued parameter to the generator.
    virtual Status setParam(const string& name, const double& value) {
      if (_initialized) {
        cerr << "Can't set parameter " << name << ": generator has already been initialized." << endl;
        throw (runtime_error("Tried to set param after generator initialization."));
      }
      return setParam(name, toString(value));
    }

    /// Pass a double/float-valued parameter to the generator.
    virtual Status setParam(const string& name, const float& value) {
      if (_initialized) {
        cerr << "Can't set parameter " << name << ": generator has already been initialized." << endl;
        throw (runtime_error("Tried to set param after generator initialization."));
      }
      return setParam(name, toString(value));
    }
    
    /// Pass a bool-valued parameter to the generator.
    virtual Status setParam(const string& name, const bool& value) {
      if (_initialized) {
        cerr << "Can't set parameter " << name << ": generator has already been initialized." << endl;
        throw (runtime_error("Tried to set param after generator initialization."));
      }
      return setParam(name, toString(value));
    }
    
    /// Pass a string-valued parameter to the generator.
    virtual Status setParam(const string& name, const string& value) {
      if (_initialized) {
        cerr << "Can't set parameter " << name << ": generator has already been initialized." << endl;
        throw (runtime_error("Tried to set param after generator initialization."));
      }
      _myState.setParam(name, value);

      return SUCCESS;
    }

    /// Run one event.
    virtual Status makeEvent(HepMC::GenEvent& evt) {
      if (!_initialized) initialize();
      clearEvent(evt);
      _myState.addEvent();
      return SUCCESS;
    }
    
    /// Return the generated cross section in pb.
    
    virtual const double getCrossSection() {
      cerr << " Have not implemented getCrossSection in generator "
      <<getName()<<endl;
      throw runtime_error
      ("Tried to access cross section in Generator base class");
      return _crossSection;
    }

    /// Finalise.
    virtual Status finalize() = 0;

  protected:
    /// Initialise.
    virtual Status initialize() {
      if (!_initialstateset) setInitialState();
      return SUCCESS;
    }

    virtual Status setGenSpecificInitialState
    (int p1 = PROTON, double e1 = 7000.0, 
     int p2 = PROTON, double e2 = 7000.0)=0;
    
    /// Clear the supplied event.
    virtual Status clearEvent(HepMC::GenEvent& evt) {
      evt.clear(); // < Requires HepMC >= 2.01.00
      return SUCCESS;
    }
    
    double _crossSection;

    /// Flag for detecting whether the setInitialState method has run.
    bool _initialstateset;

    /// Flag for detecting whether the initialization method has run.
    bool _initialized;

    /// Collection of translated names for particles
    map<PdgCode, string> _particleNames;

  private:
    
    GeneratorState _myState;
    bool _versionSet;
    string _myVersion;
  };


  ////////////////////////////////////////////////

  /// Class factory for dynamic generator loading.
  typedef Generator* (*Creator)();
  //typedef Generator* (*genbuilder_fn)();

  /// Class destroyer for dynamic generator loading.
  typedef void Destroyer(Generator*);

}

#endif
